"""Kraky websocket module"""
import asyncio
import json
import socket

import websockets

from .log import get_module_logger


class KrakyWsError(Exception):
    """Raise exception when Kraken Websocket raise an error"""


class KrakyWsClient:
    """Kraken Websocket client implementation"""

    def __init__(self, connection_env="production"):
        self.connection_env = connection_env
        self.connections = {}
        self.logger = get_module_logger(__name__)

    async def connect(self, handler, connection_name="main"):
        if self.connection_env == "production":
            ws_url = "wss://ws.kraken.com"
        elif self.connection_env == "production-auth":
            ws_url = "wss://ws-auth.kraken.com"
        elif self.connection_env == "beta":
            ws_url = "wss://beta-ws.kraken.com"
        elif self.connection_env == "beta-auth":
            ws_url = "wss://beta-ws-auth.kraken.com"
        websocket = await websockets.connect(ws_url)
        self.connections[connection_name] = websocket
        while True:
            try:
                if not websocket.open:
                    websocket = await websockets.connect(ws_url)
                    self.connections[connection_name] = websocket
                else:
                    message = await websocket.recv()
                    if "errorMessage" in message:
                        error = json.loads(message)
                        self.logger.error(error["errorMessage"])
                    else:
                        data = json.loads(message)
                        await handler(data)
            except socket.gaierror:
                self.logger.error("Connection to Kraken WS closed, reconnecting...")
                continue
            except websockets.exceptions.ConnectionClosedError:
                self.logger.error("Connection to Kraken WS closed, reconnecting...")
                continue
            except websockets.exceptions.ConnectionClosedOK:
                self.logger.error("Connection to Kraken WS closed, reconnecting...")
                continue

    async def disconnect(self, connection_name="main"):
        if self.connections[connection_name] is not None:
            await self.connections[connection_name].close()
            del self.connections[connection_name]

    async def _sub_unsub(
        self, sub_type, subscription, pairs=None, connection_name="main"
    ):
        while connection_name not in self.connections:
            await asyncio.sleep(1)
        websocket = self.connections[connection_name]
        payload = {
            "event": sub_type,
            "subscription": subscription,
        }
        if pairs:
            payload["pair"] = pairs
        await websocket.send(json.dumps(payload))

    async def subscribe(self, subscription, pairs=None, connection_name="main"):
        await self._sub_unsub("subscribe", subscription, pairs, connection_name)

    async def unsubscribe(self, subscription, pairs=None, connection_name="main"):
        await self._sub_unsub("unsubscribe", subscription, pairs, connection_name)



"""Kraky API module for Kraken's API implementation"""
import base64
import hashlib
import hmac
import urllib
import time

import httpx

from .log import get_module_logger


class KrakyApiError(Exception):
    """Raise an exception when Kraken's API raise an error"""


class KrakyApiClient:
    """Kraken API client implementation"""

    def __init__(self, api_key="", secret=""):
        self.base_url = "https://api.kraken.com"
        self.api_key = api_key
        self.secret = secret
        self.logger = get_module_logger(__name__)

    async def _request(self, uri, headers=None, data=None):
        async with httpx.AsyncClient() as client:
            result = await client.post(uri, headers=headers, data=data)
            if result.status_code not in (200, 201, 202):
                self.logger.exception(result.raise_for_status())
            # check for error
            if len(result.json()["error"]) > 0:
                raise KrakyApiError(result.json()["error"])
            return result.json()["result"]

    async def public_request(self, endpoint, data=None):
        uri = f"{self.base_url}/0/public/{endpoint}"
        return await self._request(uri, data=data)

    async def private_request(self, endpoint, data=None):
        if not data:
            data = {}
        data["nonce"] = int(time.time() * 1000)
        api_path = f"/0/private/{endpoint}"
        uri = f"{self.base_url}{api_path}"
        headers = {
            "API-Key": self.api_key,
            "API-Sign": self._sign_message(api_path, data),
        }
        return await self._request(uri, headers=headers, data=data)

    def _sign_message(self, api_path, data):
        post_data = urllib.parse.urlencode(data)
        encoded = (str(data["nonce"]) + post_data).encode()
        message = api_path.encode() + hashlib.sha256(encoded).digest()
        signature = hmac.new(base64.b64decode(self.secret), message, hashlib.sha512)
        sig_digest = base64.b64encode(signature.digest())

        return sig_digest.decode()

    async def add_order(self, data):
        return await self.private_request(endpoint="AddOrder", data=data)

    async def cancel_order(self, data):
        return await self.private_request(endpoint="CancelOrder", data=data)

    async def get_ws_token(self):
        """https://www.kraken.com/features/api#ws-auth"""
        result = await self.private_request(endpoint="GetWebSocketsToken")
        return result["token"]

    async def get_trade_balances(self):
        return await self.private_request(endpoint="Balance")

    async def get_open_positions(self):
        return await self.private_request(endpoint="OpenPositions")

    async def get_open_orders(self):
        return await self.private_request(endpoint="OpenOrders")

    async def cancel_open_order(self, txid):
        return await self.private_request(endpoint="CancelOrder", data={"txid": txid})

    async def get_trades_history(
        self, trade_type=None, trades=None, start=None, end=None, ofs=None
    ):
        data = {
            arg: value
            for arg, value in locals().items()
            if arg != "self" and value is not None
        }
        return await self.private_request(endpoint="TradesHistory", data=data)

    async def get_recent_trades(self, pair: str, since=None):
        data = {
            arg: value
            for arg, value in locals().items()
            if arg != "self" and value is not None
        }
        return await self.public_request(endpoint="Trades", data=data)

    async def get_ohlc_data(self, pair: str, interval: int, since=None):
        data = {
            arg: value
            for arg, value in locals().items()
            if arg != "self" and value is not None
        }
        return await self.public_request(endpoint="OHLC", data=data)

    async def get_asset_pairs(self, info=None, pair=None):
        data = {
            arg: value
            for arg, value in locals().items()
            if arg != "self" and value is not None
        }
        return await self.public_request(endpoint="AssetPairs", data=data)

    async def get_last_price(self, pair: str):
        ohlc = await self.get_ohlc_data(pair, interval=1)
        return float(list(ohlc.values())[0][-1][4])
